package com.tzld.longarticle.recommend.server.service.recommend.score;


import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.tzld.longarticle.recommend.server.common.CostMonitor;
import com.tzld.longarticle.recommend.server.common.ThreadPoolFactory;
import com.tzld.longarticle.recommend.server.common.enums.recommend.RankStrategyEnum;
import com.tzld.longarticle.recommend.server.common.enums.recommend.ScoreStrategyEnum;
import com.tzld.longarticle.recommend.server.util.CommonCollectionUtils;
import com.tzld.longarticle.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeansException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Service;

import javax.annotation.PostConstruct;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

/**
 * @author dyp
 */
@Service
@Slf4j
public class ScoreService implements ApplicationContextAware {

    private final Map<String, ScoreStrategy> strategyMap = new HashMap<>();
    private ApplicationContext applicationContext;
    private final ExecutorService pool = ThreadPoolFactory.scorePool();

    @ApolloJsonValue("${similarity.stop.strategy.flag:[]}")
    private List<String> similarityStopStrategies;

    @PostConstruct
    public void init() {
        Map<String, ScoreStrategy> type = applicationContext.getBeansOfType(ScoreStrategy.class);
        for (Map.Entry<String, ScoreStrategy> entry : type.entrySet()) {
            ScoreStrategy value = entry.getValue();
            strategyMap.put(value.getClass().getSimpleName(), value);
        }
    }

    public ScoreResult score(ScoreParam param) {
        long start = System.currentTimeMillis();
        List<ScoreStrategy> strategies = getScoreStrategy(param);
        log.info("ScoreStrategy {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies,
                s -> s.getClass().getSimpleName())));
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<List<Score>>> futures = new ArrayList<>();
        for (final ScoreStrategy strategy : strategies) {
            Future<List<Score>> future = pool.submit(() -> {
                try {
                    long t1 = System.currentTimeMillis();
                    List<Score> result = strategy.score(param);
                    long t2 = System.currentTimeMillis();
                    CostMonitor.logCost("Score", strategy.getClass().getSimpleName(), t2 - t1);
                    return result;
                } catch (Exception e) {
                    log.error("scoreService error:{}", e.getMessage(), e);
                } finally {
                    cdl.countDown();
                }
                return new ArrayList<>();
            });
            futures.add(future);
        }
        try {
            cdl.await(30000, TimeUnit.MILLISECONDS);
        } catch (InterruptedException e) {
            log.error("rank error", e);
            return null;
        }

        Map<String, Map<String, Double>> scoreMap = new HashMap<>();
        for (Future<List<Score>> f : futures) {
            try {
                List<Score> data = f.get();
                for (Score score : data) {
                    Map<String, Double> map
                            = scoreMap.computeIfAbsent(score.getContentId(), k -> new HashMap<>());
                    map.put(score.getStrategy(), score.getScore());
                }
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }
        log.info("score耗时：{}", System.currentTimeMillis() - start);
        return new ScoreResult(scoreMap);
    }

    private List<ScoreStrategy> getScoreStrategy(ScoreParam param) {
        List<ScoreStrategy> strategies = new ArrayList<>();
        if (StringUtils.equals(param.getStrategy(), RankStrategyEnum.RANDOM_STRATEGY.getStrategy())) {
            return strategies;
        }
        strategies.add(strategyMap.get(ScoreStrategyEnum.CRAWLER_DAYS_DECREASE_STRATEGY.value()));
        if (!similarityStopStrategies.contains(param.getStrategy())
                && !StringUtils.equals(param.getStrategy(), RankStrategyEnum.HIS_JUMP_STRATEGY.getStrategy())) {
            strategies.add(strategyMap.get(ScoreStrategyEnum.SIMILARITY.value()));
        }
        if (StringUtils.equals(param.getStrategy(), RankStrategyEnum.LATE_STRATEGY.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.INFINITE_STRATEGY.getStrategy())) {
            strategies.add(strategyMap.get(ScoreStrategyEnum.CATEGORY.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.ACCOUNT_PRE_DISTRIBUTE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.PUBLISH_TIMES.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.FLOW_CTL_DECREASE.value()));
            return strategies;
        }
        strategies.add(strategyMap.get(ScoreStrategyEnum.VIEW_COUNT.value()));
        if (StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV3.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV4.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.HIS_JUMP_STRATEGY.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV5.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV7.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV8.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV9.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV10.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV11.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV12.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV13.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV14.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV15.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV16.getStrategy())
                || StringUtils.equals(param.getStrategy(), RankStrategyEnum.ArticleRankV17.getStrategy())) {
            strategies.add(strategyMap.get(ScoreStrategyEnum.CATEGORY.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.ACCOUNT_PRE_DISTRIBUTE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.FLOW_CTL_DECREASE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.VIEW_COUNT_RATE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.VIEW_COUNT_RATE_V2.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.VIEW_COUNT_RATE_CORRELATION.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.PUBLISH_TIMES.value()));
//            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_FANS_RATE_RATE.value()));
//            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_FANS_SUM_RATE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_AVG_READ_RATE_RATE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_AVG_READ_RATE_CORRELATION_RATE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_AVG_READ_SUM_RATE.value()));
            strategies.add(strategyMap.get(ScoreStrategyEnum.HIS_FISSION_DE_WEIGHT_AVG_READ_SUM_RATE.value()));
        }

        return strategies;
    }

    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }

}
